/***********************************************************************
Viber: parallel 1D wave equation solver

Copyright (C) 2007 Andrea Sottoriva

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <mpi.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define WINSIZE 600

#define DEBUG

#ifdef DEBUG
#define dprintf(format, args...) fprintf(stdout, format , ## args)
#else
#define dprintf (void)
#endif

#define SIZE_PARMS 0
#define STR_INIT   1
#define STR_SYNC   2
#define LEFT_SYNC  3
#define RIGHT_SYNC 4
#define STR_GATHER 5

#define TRUE  1
#define FALSE 0

float *stringa[3];
float c;
float dt;
float dx;
float tau;
float damp;
float total_t = 0.f;
int delay = 0;
int cur = 1, prev = 0, next = 2;

/* visualization stuff */
Display  *mydisp;
Window   mywin;
GC       mygc;
XColor   col;
XEvent   myevent;
KeySym   mykey;
XSizeHints myhint;
XPointer  arg;
Colormap     cmap;
int      myscreen;
int      xvmin, yvmin, xvmax, yvmax;
unsigned long   myfore, myback;

/* initialize an X11 window */
void initX11(void) 
{
  if((mydisp = XOpenDisplay("")) == NULL) {
    fprintf(stderr, "Cannot open display\n");
    MPI_Finalize();
    exit(EXIT_FAILURE);
  }

  cmap=DefaultColormap(mydisp, 0);
  myscreen = DefaultScreen(mydisp);
  myfore = WhitePixel(mydisp, myscreen);
  myback = BlackPixel(mydisp, myscreen);
  xvmin = myhint.x = 10;
  yvmin = myhint.y = 10;
  xvmax = myhint.width = myhint.min_width = myhint.max_width = WINSIZE;
  yvmax = myhint.height = myhint.min_height = myhint.max_height = WINSIZE;
  myhint.flags = USPosition | PSize | PMinSize | PMaxSize;
  mywin = XCreateSimpleWindow(mydisp,DefaultRootWindow(mydisp), myhint.x, myhint.y,
			      myhint.width,myhint.height,5,myfore,myback);
  XSetStandardProperties(mydisp,mywin,"Viber - Andrea Sottoriva","Viber - Andrea Sottoriva",None,None,None,&myhint);
  mygc = XCreateGC(mydisp,mywin,0,0);
  XSetBackground(mydisp,mygc,myback);
  XSetForeground(mydisp,mygc,myfore);
  XSelectInput(mydisp,mywin,ButtonPressMask | KeyPressMask | ExposureMask | Button1MotionMask | Button3MotionMask);

  XMapRaised(mydisp,mywin);
  XClearWindow(mydisp, mywin);
}

double sech(double x)
{
  return 2. / (exp(x) + exp(-x));
}

/* 
   Abort program
 */
void abortMPI(int err, int rank)
{
  char error_str[MPI_MAX_ERROR_STRING];
  int len;
  int init_flag;

  MPI_Error_string(err, error_str, &len);

  if(rank < 0) {
    fprintf(stderr, "Error from %d: %s (errno = %d)\n", rank, error_str, err);
  }
  else {
    fprintf(stderr, "Global error: %s (errno = %d)\n", error_str, err);
  }

  MPI_Initialized(&init_flag);

  if(init_flag)
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);

  exit(EXIT_FAILURE);
} 

void usage(void)
{
  fprintf(stderr, "\n\nUsage: ./viber <string_length> <number_of_steps> [-c C] [-t dt] [-x dx] [-d damp] [-display] [-s frame_delay(us)] [-periodic]\n\n");
  MPI_Finalize();
  exit(EXIT_FAILURE);
}

Bool checkKeyEvent(Display* dpy, XEvent* event, XPointer arg)
{
  return event->type == KeyPress;
}

int main(int argc, char** argv)
{
  char proc_name[MPI_MAX_PROCESSOR_NAME];
  int ret, nproc, length, i, j, size, npoints, nsteps, offset, rank = -1;
  int int_params[3];
  float float_params[4];
  int period = FALSE;
  int display = FALSE;
  MPI_Status status;
  MPI_Comm grid1d;

  /* init MPI */
  if((ret = MPI_Init(&argc, &argv)))
    abortMPI(ret, rank);

  /* get task rank */
  if((ret = MPI_Comm_rank(MPI_COMM_WORLD, &rank)))
    abortMPI(ret, rank);

  /* get total number of tasks */
  if((ret = MPI_Comm_size(MPI_COMM_WORLD, &nproc)))
    abortMPI(ret, rank);
  
  /* get processor name */
  if(MPI_Get_processor_name(proc_name, &length))
    abortMPI(ret, rank);

  /* command line parsing */
  if(rank == 0) {

    c = 0.04f;
    dt = 0.1f;
    dx = 0.02f;
    damp = 0.0f;
    
    if(argc < 3) {
      usage();
    }
    
    size = atoi(argv[1]);
    nsteps = atoi(argv[2]);
    
    for(i = 3; i < argc; ++i) {
      if(strcmp(argv[i], "-c") == 0) {
	if(i == argc - 1) usage();
	else {
	  c = atof(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-t") == 0) {
	if(i == argc - 1) usage();
	else {
	  dt = atof(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-x") == 0) {
	if(i == argc - 1) usage();
	else {
	  dx = atof(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-d") == 0) {
	if(i == argc - 1) usage();
	else {
	  damp = atof(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-s") == 0) {
	if(i == argc - 1) usage();
	else {
	  delay = atoi(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-periodic") == 0) {
	period = TRUE;
      }
      else if(strcmp(argv[i], "-display") == 0) {
	display = TRUE;
      }
    }  

    for(i = 0; i < nproc; ++i)
      if((ret = MPI_Send(&period, 1, MPI_INT, i, SIZE_PARMS, MPI_COMM_WORLD)))
	abortMPI(ret, rank);      
  }
  else
    if((ret = MPI_Recv(&period, 1, MPI_INT, 0, SIZE_PARMS, MPI_COMM_WORLD, &status)))
      abortMPI(ret, rank);

  /* create a 1D cartesian grid topology */
  if((ret = MPI_Cart_create(MPI_COMM_WORLD, 1, &nproc, &period, FALSE, &grid1d)))
    abortMPI(ret, rank);

  dprintf("Viber on %*s, %d\n", length, proc_name, rank);

  /* get the new task rank referring to the topology */
  if((ret = MPI_Comm_rank(grid1d, &rank)))
    abortMPI(ret, rank);

  if(rank == 0) {
    
    stringa[prev] = (float*)malloc(size * sizeof(float));
    stringa[cur] = (float*)malloc(size * sizeof(float));
    stringa[next] = (float*)malloc(size * sizeof(float));

    dprintf("Initializing processes:\tsize = %d\n\t\tsteps = %d\n\t\tc = %.2f\n\t\t\tdt = %.2f\n\t\t\tdx = %.2f\n\t\t\tdamp = %.2f\n", size, nsteps, c, dt, dx, damp);

    /* initializing the string */
    for(i = 0; i < size; ++i) {
      stringa[cur][i] = stringa[prev][i] = 2.5 * sech((double)(i - size / 2 + 20) / 5.f) * sin(i * 10.f * M_PI / (size - 1) + 10);
      //stringa[cur][i] = stringa[prev][i] = sin(i * 2 * M_PI  / (size - 1));
    }
    stringa[cur][i - 1] = stringa[prev][i - 1] = stringa[cur][0] = stringa[prev][0] = 0.f;

    /* delivery parameters */
    for(i = 1; i < nproc; ++i) {

      /* domain decomposition */
      npoints = (int)(size / nproc + (size % nproc + nproc - i - 1) / nproc);
      if(npoints > (int)(size / nproc))
	offset = npoints * i;
      else
	offset = size - (nproc - i) * npoints;

      dprintf("Deliverying params [ npoints = %d, offset = %d ] to process %d\n", npoints, offset, i);

      int_params[0] = npoints;
      int_params[1] = nsteps;
      int_params[2] = display;
      float_params[0] = c;
      float_params[1] = dt;
      float_params[2] = dx;
      float_params[3] = damp;

      /* send parameters */
      if((ret = MPI_Send(int_params, 3, MPI_INT, i, SIZE_PARMS, grid1d)))
	abortMPI(ret, rank);

      if((ret = MPI_Send(float_params, 4, MPI_FLOAT, i, SIZE_PARMS, grid1d)))
	abortMPI(ret, rank);

      /* send the string */
      if((ret = MPI_Send(&stringa[cur][offset], npoints, MPI_FLOAT, i, STR_INIT, grid1d)))
	abortMPI(ret, rank);
    }

    /* domain decomposition for the root process */
    npoints = (int)(size / nproc + (size % nproc + nproc - 1) / nproc);    
    if(npoints > (int)(size / nproc))
      offset = 0;
    else
      offset = size - nproc * npoints;

    dprintf("Root process [ npoints = %d, offset = %d ]\n", npoints, offset);

    /* open a window */
    if(display)
      initX11();
  }
  else {

    /* receive parameters */
    if((ret = MPI_Recv(int_params, 3, MPI_INT, 0, SIZE_PARMS, grid1d, &status)))
      abortMPI(ret, rank);

    npoints = int_params[0];
    nsteps = int_params[1];
    display = int_params[2];

    if((ret = MPI_Recv(float_params, 4, MPI_FLOAT, 0, SIZE_PARMS, grid1d, &status)))
      abortMPI(ret, rank);

    c = float_params[0];
    dt = float_params[1];
    dx = float_params[2];
    damp = float_params[3];

    stringa[prev] = (float*)malloc(npoints * sizeof(float));
    stringa[cur] = (float*)malloc(npoints * sizeof(float));
    stringa[next] = (float*)malloc(npoints * sizeof(float));

    /* receive my own part of the string */
    if((ret = MPI_Recv(stringa[cur], npoints, MPI_FLOAT, 0, STR_INIT, grid1d, &status)))
      abortMPI(ret, rank);

    memcpy(stringa[prev], stringa[cur], npoints * sizeof(float));
  }
  
  int l_rank, r_rank;
  float left, right;

  left = right = 0.0f;

  /* get the rank of the two neighbours */
  if((ret = MPI_Cart_shift(grid1d, 0, 1, &l_rank, &r_rank)))
    abortMPI(ret, rank);

  dprintf("Process %d: left rank = %d, right rank = %d\n", rank, l_rank, r_rank);

  tau = c * (dt / dx);

  float factor = (float)WINSIZE / (size + 2);
  float amp = factor * 5;
  int cl_npoints, cl_offset;

  /* compute the string oscillations */
  for(i = 0; i < nsteps; ++i) {

    if(display && rank == 0 && XCheckIfEvent(mydisp, &myevent, checkKeyEvent, arg) && myevent.xkey.keycode == 9) {
      XFreeGC(mydisp,mygc);
      XDestroyWindow(mydisp,mywin);
      XCloseDisplay(mydisp);
      
      if((ret = MPI_Finalize()))
	abortMPI(ret, rank);
      
      exit(EXIT_SUCCESS);
    }

    if(nproc > 1) {

      /* synchronization */
      if(rank % 2 == 0) {
	
	if(r_rank >= 0) {
	  //dprintf("%d: right syncronizing with %d\n", rank, r_rank);
	  if((ret = MPI_Sendrecv(&stringa[cur][npoints - 1], 1, MPI_FLOAT, r_rank, LEFT_SYNC,
				 &right, 1, MPI_FLOAT, r_rank, RIGHT_SYNC, grid1d, &status)))
	    abortMPI(ret, rank);
	}
	
	if(l_rank >= 0) {
	  //dprintf("%d: left syncronizing with %d\n", rank, l_rank);
	  if((ret = MPI_Sendrecv(&stringa[cur][0], 1, MPI_FLOAT, l_rank, RIGHT_SYNC,
				 &left, 1, MPI_FLOAT, l_rank, LEFT_SYNC, grid1d, &status)))
	    abortMPI(ret, rank);
	}
      }
      else {
	if(l_rank >= 0) {
	  //dprintf("%d: left syncronizing with %d\n", rank, l_rank);
	  if((ret = MPI_Sendrecv(&stringa[cur][0], 1, MPI_FLOAT, l_rank, RIGHT_SYNC,
				 &left, 1, MPI_FLOAT, l_rank, LEFT_SYNC, grid1d, &status)))
	    abortMPI(ret, rank);
	}
	if(r_rank >= 0) {
	  //dprintf("%d: right syncronizing with %d\n", rank, r_rank);
	  if((ret = MPI_Sendrecv(&stringa[cur][npoints - 1], 1, MPI_FLOAT, r_rank, LEFT_SYNC,
				 &right, 1, MPI_FLOAT, r_rank, RIGHT_SYNC, grid1d, &status)))
	    abortMPI(ret, rank);
	}
      }
    }

    /* compute the wave equation locally */
    for(j = 0; j < npoints; ++j) {

      stringa[next][j] =
	2 * stringa[cur][j] - stringa[prev][j] +
	tau * tau * (
		     ((j == 0) ? left : stringa[cur][j - 1]) -
		     2 * stringa[cur][j] +
		     ((j == npoints - 1) ? right : stringa[cur][j + 1])) -
	damp * dt * (stringa[cur][j] - stringa[prev][j]);
    }

    /* switch indexes */
    prev = (prev + 1) % 3;
    cur = (cur + 1) % 3;
    next = (next + 1) % 3;
    
    /* collect all the string tokens */
    if(display) {

      if(rank == 0) {
	
	char label[128];
	
	/* update total time counter */
	total_t += dt;
	
	XSetForeground(mydisp, mygc, 128);
	
	for(j = 1; j < nproc; ++j) {
	  cl_npoints = (int)(size / nproc + (size % nproc + nproc - j - 1) / nproc);
	  
	  if(cl_npoints > (int)(size / nproc))
	    cl_offset = cl_npoints * j;
	  else
	    cl_offset = size - (nproc - j) * cl_npoints;
	  
	  //dprintf("Getting the string [ %d, %d ] from %d\n", cl_npoints, cl_offset, j);
	  
	  if((ret = MPI_Recv(&stringa[cur][cl_offset], cl_npoints, MPI_FLOAT, j, STR_GATHER, grid1d, &status)))
	    abortMPI(ret, rank);
	  
	  XDrawLine(mydisp, mywin, mygc, cl_offset * factor, 0, cl_offset * factor, WINSIZE);
	}
	
	XSetForeground(mydisp, mygc, myfore);
	
	if(period == TRUE)
	  left = right = (stringa[cur][0] + stringa[cur][npoints - 1]) / 2;
	else
	  left = right = 0.f;
	
	/* draw the string */
	XDrawLine(mydisp, mywin, mygc,
		  0, left * factor * amp + WINSIZE / 2,
		  factor, stringa[cur][0] * factor * amp + WINSIZE / 2);
	for(j = 0; j < size - 1; ++j) {
	  XDrawLine(mydisp, mywin, mygc,
		    (j + 1) * factor, stringa[cur][j] * factor * amp + WINSIZE / 2,
		    (j + 2) * factor, stringa[cur][j + 1] * factor * amp + WINSIZE / 2);
	}
	XDrawLine(mydisp, mywin, mygc,
		  size * factor, stringa[cur][size - 1] * factor * amp + WINSIZE / 2,
		  (size + 1) * factor, right * factor * amp + WINSIZE / 2);
	
	snprintf(label, 120, "Viber: %d processors, time = %.2f s", nproc, total_t);
	XDrawImageString(mydisp, mywin, mygc, 2, 15, label, strlen(label));
	
	XFlush(mydisp);
	XClearWindow(mydisp, mywin);
	usleep(delay);
      }
      else {	/* send the string to the root process */
	if((ret = MPI_Send(stringa[cur], npoints, MPI_FLOAT, 0, STR_GATHER, grid1d)))
	  abortMPI(ret, rank);
      }
    }
  }

  /* sync and die */
  if((ret = MPI_Barrier(grid1d)))
    abortMPI(ret, rank);

  if((ret = MPI_Finalize()))
    abortMPI(ret, rank);

  if(display && rank == 0) {
    XFreeGC(mydisp,mygc);
    XDestroyWindow(mydisp,mywin);
    XCloseDisplay(mydisp);  
  }

  exit(EXIT_SUCCESS);
}   
